use "ARITHMETIC_MACHINE.sig";
(* This is the machine we just looked at in the main file. *)
structure StraightforwardMachine :> ARITHMETIC_MACHINE =
struct
  datatype exp = Constant of int
               | Negate of exp
               | Add of exp * exp
               | Subtract of exp * exp
               | Multiply of exp * exp
  
  datatype oper = Const of int
                | Neg
                | Plus
                | Minus
                | Times

  type code = oper list

  fun compile e =
    case e of
      Constant i => [Const i]
    | Negate e => (compile e) @ [Neg]
    | Add (e1, e2) => (compile e1) @ (compile e2) @ [Plus]
    | Subtract (e1, e2) => (compile e1) @ (compile e2) @ [Minus]
    | Multiply (e1, e2) => (compile e1) @ (compile e2) @ [Times]

  type state = int list
  exception BadInput;

  fun unary f stack =
    case stack of
      e::tail => (f e)::tail
    | _ => raise BadInput

  fun binary f stack =
    case stack of
      r::l::tail => (f (l, r))::tail
    | _ => raise BadInput

  fun step oper =
    case oper of
      Const i => (fn s => i::s)
    | Neg => unary (op~)
    | Plus => binary (op+)
    | Minus => binary (op-)
    | Times => binary (op*)

  type value = int
  fun extract s =
    case s of
      [v] => v
    | _ => raise BadInput
end
